
#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/random.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/cache.h>
#include <linux/sysctl.h>
#include <linux/proc_fs.h>
#include <linux/if_arp.h>

#include <asm/uaccess.h>


// Expand SKB's headroom and tailroom by HEAD_NEED and TAIL_NEED amount. 
// Return the modified skb upon success and NULL, otherwise. 
extern struct sk_buff* 
expand_skb(struct sk_buff *skb, int head_need, int tail_need) 
{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0))
	if (head_need > skb_headroom(skb) || tail_need > skb_tailroom(skb)) {
		struct sk_buff *nskb;

		nskb = dev_alloc_skb(skb->len + head_need + tail_need);
		if (nskb == NULL) {
			//dev_kfree_skb_any(skb);
			return NULL;
		}
		
		skb_reserve(nskb, head_need);
		memcpy(skb_put(nskb, skb->len), skb->data, skb->len);
		dev_kfree_skb_any(skb);
		return nskb;
	}
	return skb; // has enough tail and head rooms
	
#else // Linux 2.4.0 or newer 
	
	head_need -= skb_headroom(skb);
	tail_need -= skb_tailroom(skb);
	
	if (head_need > 0 || tail_need > 0) {
		if (pskb_expand_head(skb, head_need > 0 ? head_need : 0,
				     tail_need > 0 ? tail_need : 0,
				     GFP_ATOMIC)) {
			printk(KERN_DEBUG "%ld expand_skb failed to " 
			       "reallocate skb buffer\n", jiffies);
			return NULL;
		}
		//printk(KERN_DEBUG "%ld expand_skb success! But consider per-allocating buffer!\n", jiffies);
	}
	return skb;
#endif
}



#define BEGIN_SYS_CALL() \
        mm_segment_t fs; \
	fs = get_fs();  \
	set_fs (get_ds())
 

#define END_SYS_CALL()  set_fs(fs)


// Return timeval that corresponds to TV1 - TV2. 
struct timeval time_diff(const struct timeval *tv1, const struct timeval *tv2) 
{
	struct timeval tv; 
	int carry = 0;
	tv.tv_usec = tv1->tv_usec - tv2->tv_usec;
	if (tv.tv_usec < 0) {
		tv.tv_usec += 1000000;
		carry = -1;
	}
	
	tv.tv_sec = tv1->tv_sec - tv2->tv_sec + carry;
	if (tv.tv_sec < 0 && carry == -1) // make usec negative as well
		tv.tv_usec = - tv.tv_usec;
	return tv;
}
